/*
DL2 Monitor: ./pa/papppoe.c
Copyright (C) 2002, Rolf Jger, Ulm (Germany)

See the file README for information on usage and redistribution
of this source code file, and for a DISCLAIMER OF ALL WARRANTIES.
*/

/*----------------------------------------------------------------------------
 * Datalink Monitor (papppoe): Protocol Analysis (RFC2516)
 * PPP over Ethernet Framing
 *
 * This module analyzes the PPPoE encapsulation of PPP. It recognizes some
 * fields of the PPPoE packet and interpretes the PPP protocol field of the
 * PPPoE payload. If requested, it calls an analyzer for the recognized PPP
 * protocol. Currently supported protocols are: lcp, ipcp, pap, chap, and ip.
 *
 * Implementation Notes
 * P_* indicate frame positions of certain fields;
 * C_* indicate known values of the code field;
 * default contains: code/discovery, payload-type
 * 
 * Author: Rolf Jger,Ulm
 *----------------------------------------------------------------------------
 * YYYYMMDD  History
 * 20020410  Mdefault: only solved for discovery stage.
 * 20020330  created from scratch
 * $Id: papppoe.c,v 1.3 2002/04/14 21:07:54 dv Exp $
 *----------------------------------------------------------------------------
*/
# include <stdio.h>
# include <string.h>
# include "lib/dl2pa.h"
# include "lib/paether.h"
# include "lib/pappp.h"
# include "pa/pppoe.h"

//-- common to any PPPoE payload ---------------------------------------------
# define P_PVT  0x00
# define P_CODE  0x01
# define P_SESSIONID  0x02
# define P_LENGTH  0x04
# define P_PAYLOAD  0x06
# define P_DATA  0x08

# define C_PADI  0x09
# define C_PADO  0x07
# define C_PADR  0x19
# define C_PADS  0x65
# define C_PADT  0xa7
# define C_SESS  0x00

# define CI_PADI   "PADI"
# define CI_PADO   "PADO"
# define CI_PADR   "PADR"
# define CI_PADS   "PADS"
# define CI_PADT   "PADT"
# define CI_SESS   "SESS"


//-- common to any PPP frame -------------------------------------------------
typedef struct Pppinfo {
  Doctet type; 
  const char *name;
  int pid;
  int (*analyzer)(Dl2_Frame*,int);
} Pppinfo;

# define PPPINFO_ADD(t,a) {PAPPP_T_##t,#t,PA_PID_##t,a}
static const Pppinfo pppinfo[] = {
  PPPINFO_ADD(lcp,dl2_palcp),
  PPPINFO_ADD(pap,dl2_papap),
  PPPINFO_ADD(lqr,0),           //not supported
  PPPINFO_ADD(chap,dl2_pachap), 
  PPPINFO_ADD(ipcp,dl2_paipcp), 
  PPPINFO_ADD(ip,dl2_paip),   
  PPPINFO_ADD(ctcp,0),          //not supported
  PPPINFO_ADD(utcp,0),          //not supported
  {0,NULL,0} 
};

static const Pppinfo* get_pppinfo(Doctet);

//-exports--------------------------------------------------------------------
int dl2_papppoe(Dl2_Frame *frame,int verbose) {
  Octet *buffer = frame->buffer;
  int length = frame->length, i, pl;
  Octet code;
  Doctet protocol;
  const Pppinfo *pi;

  // check for packet length:
  if (length < P_PAYLOAD) return -1;

  // protocol version/type info:
  if (verbose & PPPOE_Mprotocol_version) {
    printf(" %s=",PPPOE_Fprotocol_version);
    for (i = P_PVT; i < P_CODE; i++) printf("%02x",buffer[i]);
  } else i = P_CODE;

  // code:
  code = buffer[i];
  if (verbose & PPPOE_Mcode) {
    printf(" %s=",PPPOE_Fcode);
    for (; i < P_SESSIONID; i++) printf("%02x",buffer[i]);
  }
  else if (verbose & PPPOE_Mdefault) {
    const char *codestring = PA_FC_UNKNOWN;
    if (frame->type == PAETH_T_PPPoE_D) { 
      switch (code) {
      case C_PADI:
        codestring = CI_PADI; break;
      case C_PADO:
        codestring = CI_PADO; break;
      case C_PADR:
        codestring = CI_PADR; break;
      case C_PADS:
        codestring = CI_PADS; break;
      case C_PADT:
        codestring = CI_PADT; break;
      }
      printf(" %s",codestring);
    }
  } i = P_SESSIONID;

  // session id:
  if (verbose & PPPOE_Msession_id) {
    printf(" %s=",PPPOE_Fsession_id);
    for (; i < P_LENGTH; i++) printf("%02x",buffer[i]);
  } else i = P_LENGTH;

  // length:
  pl = str2doct(buffer+i);
  if (verbose & PPPOE_Mpayload_length) {
    printf(" %s=",PPPOE_Fpayload_length);
    for (; i < P_PAYLOAD; i++) printf("%02x",buffer[i]);
  } else i = P_PAYLOAD;

  // payload-type (PPP protocol field):
  protocol = str2doct(buffer+i);
  pi = get_pppinfo(protocol);
  if (verbose & PPPOE_Mpayload_type) {
    printf(" %s=",PPPOE_Fpayload_type);
    for (; i < P_DATA; i++) printf("%02x",buffer[i]);
  } 
  else if (verbose & PPPOE_Mdefault) {
    if (pi != NULL) printf(" %s",pi->name);
  } i = P_DATA;

  // payload-data: tbd
  
  frame->length = pl;
  frame->buffer = buffer+i;
  if (verbose & PPPOE_Mfollow_frame) {
    int j;
    if (pi != NULL && pi->analyzer) {
      printf(" [");
      j = pi->analyzer(frame,pa_verbose(pi->pid));
      printf(" ]");
      i += j;
    }
  } else i += pl;

  return i;
} //dl2_papppoe


//-locals---------------------------------------------------------------------
const Pppinfo* get_pppinfo(Doctet type) {
  const Pppinfo *pi = pppinfo;
  for (pi = pppinfo; pi->type != 0; pi++) {
    if (pi->type == type) return pi;
  } return NULL;
} //get_pppinfo

/***EOM***/
